#!/usr/bin/perl

use strict;
use warnings;
use Getopt::Long;
use Data::Dumper;
use List::MoreUtils;
use List::Util qw(max);
use Statistics::Descriptive; 
use Pod::Usage;

my %data      = ();
my %Ratios    = ();
my %stats     = ();

my $bam_fh        = '';            # <STRING>: input bam file (must be paired end data)
my $mode          = 'GeomMean';    # <STRING> <InMax|GeomMean|InLink|uniform|coverage>: specifies which ratio should be tested
my $maxFrag       = 1000;          # <INT> greater than 1: only proper pairs with a insert length less then $maxFrag are considered
my $minReads      = 1;             # <INT>: only sites are considered with >= INT reads ending/starting
my $controlSample = "read";        # <STRING> <mate|pair>: which side is the not enricht one and used as a background distribution
my $testSample    = "mate";        # <STRING> <mate|pair>: which side is the not enricht one and used as a background distribution
my $bg_fh         = "";            # <STRING>: path to file where signal used should be dumped at (in bedgraph format)
my $samtools      = "samtools";    # <STRING>: path to samtools
my $verbose       = 0;

my @time=localtime();
my $time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));


GetOptions(
  "bam=s"       => \$bam_fh,
  "mode=s"      => \$mode,
  "mf=i"        => \$maxFrag,
  "mr=i"        => \$minReads,
  "bg=s"        => \$bg_fh,
  "samtools=s"  => \$samtools,
  "v"           => \$verbose,
  "man"         => sub{pod2usage(-verbose => 2)},
  "help|?"      => sub{pod2usage(-verbose => 1)}
    );

# check if bam file exists.
unless (-f $bam_fh){
  print STDERR "Error:\tno existing bam file was specified. Please use option -bam <BAM FILE>.\n";
  exit;
}

# read in bam and save coverage of 5' ends of reads and of mates for each position
@time=localtime();
$time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));
print STDERR "Log:\tStart reading bam file $bam_fh @ $time_string\n" if ($verbose);
open BAM, "$samtools view $bam_fh |" or die "can t open <$samtools view $bam_fh | ... >\n";
while(<BAM>){
  chomp;
  my @F=split"\t",$_;
  my ($bitflag, $chr, $leftpos, $cigar, $pnext, $fraglength, $rnext) = ($F[1], $F[2], $F[3], $F[5], $F[7], $F[8], $F[6]);
  my $cigarlength = &cigarlength($cigar);
  
  next if ($pnext == 0);
  next if ($fraglength == 0);
  next if ($rnext ne "=");
  
  if( ($bitflag & 1) && ($bitflag & 2) && ($bitflag & 16) && !($bitflag & 32) && ($bitflag & 64)  ){  # $bitflag == 83
    my $readEnd = $leftpos + $cigarlength -1;
    my $mateEnd = $pnext;
    
    next if (  ($readEnd - $mateEnd) < 0 || ($readEnd - $mateEnd) > $maxFrag );
    
    $data{read}->{rev}->{$chr}->{$readEnd}->{coverage}++;
    $data{read}->{rev}->{$chr}->{$readEnd}->{in}->{$mateEnd}++;
    $data{mate}->{rev}->{$chr}->{$mateEnd}->{coverage}++;
    $data{mate}->{rev}->{$chr}->{$mateEnd}->{in}->{$readEnd}++;
  }
  elsif( ($bitflag & 1) && ($bitflag & 2) && !($bitflag & 16) && ($bitflag & 32) && ($bitflag & 64)  ){  # $bitflag == 99
    my $readEnd = $leftpos;
    my $mateEnd = $leftpos + $fraglength - 1;
    
    next if ( ($mateEnd - $readEnd) < 0  || ($mateEnd - $readEnd) > $maxFrag );
    
    $data{read}->{fwd}->{$chr}->{$readEnd}->{coverage}++;
    $data{read}->{fwd}->{$chr}->{$readEnd}->{in}->{$mateEnd}++;
    $data{mate}->{fwd}->{$chr}->{$mateEnd}->{coverage}++;
    $data{mate}->{fwd}->{$chr}->{$mateEnd}->{in}->{$readEnd}++;
  }
}
close BAM;
print STDERR "\t\t..finished\n" if ($verbose);

if ($bg_fh){
  @time=localtime();
  $time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));
  print STDERR "Log:\tStart printing bedgraph file of considered coverage signals @ $time_string\n";
  open BGfwd, "> ${bg_fh}.fwd.bg" or die "Error\tcan t write to ${bg_fh}.fwd.bg\n";
  open BGrev, "> ${bg_fh}.rev.bg" or die "Error\tcan t write to ${bg_fh}.rev.bg\n";
  
  my $type  = $testSample;
  foreach my $strand (qw/fwd/){
    foreach my $chr (sort {$a cmp $b} keys %{$data{$type}->{$strand}}){
      foreach my $pos (sort {$a <=> $b} keys %{$data{$type}->{$strand}->{$chr}}){
	next if ( $data{$type}->{$strand}->{$chr}->{$pos}->{coverage} < $minReads );
	print BGfwd join("\t", $chr, $pos-1, $pos, $data{$type}->{$strand}->{$chr}->{$pos}->{coverage})."\n";
      }
    }
  }
  foreach my $strand (qw/rev/){
    foreach my $chr (sort {$a cmp $b} keys %{$data{$type}->{$strand}}){
      foreach my $pos (sort {$a <=> $b} keys %{$data{$type}->{$strand}->{$chr}}){
	next if ( $data{$type}->{$strand}->{$chr}->{$pos}->{coverage} < $minReads );
	print BGrev join("\t", $chr, $pos-1, $pos, -1*$data{$type}->{$strand}->{$chr}->{$pos}->{coverage})."\n";
      }
    }
  }
  print STDERR "\t\t..finished\n" if ($verbose);
}


# calculate for each position a score according to the $mode specified
@time=localtime();
$time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));
print STDERR "Log:\tStart calculating ratios ($mode) @ $time_string\n" if ($verbose);
foreach my $type (keys %data){
  foreach my $strand (keys %{$data{$type}}){
    foreach my $chr (keys %{$data{$type}->{$strand}}){
      foreach my $pos (keys %{$data{$type}->{$strand}->{$chr}}){
	next if ( $data{$type}->{$strand}->{$chr}->{$pos}->{coverage} < $minReads );
        if ( $mode eq "InMax"){          # ratio at position i is coverage_i/max(coverages_j) for all j's where there is a read-mate pair between i and j
	  $Ratios{$type}->{$strand}->{$chr}->{$pos} = $data{$type}->{$strand}->{$chr}->{$pos}->{coverage}/max(values %{$data{$type}->{$strand}->{$chr}->{$pos}->{in}});
        }
        elsif( $mode eq "GeomMean"){     # ratio at position i is geometric_mean(coverage_i/coverages_j) for all j's where there is a read-mate pair between i and j
          my @values = ();
          foreach my $inPos (keys %{$data{$type}->{$strand}->{$chr}->{$pos}->{in}}){
            push @values, ($data{$type}->{$strand}->{$chr}->{$pos}->{coverage}/$data{$type}->{$strand}->{$chr}->{$pos}->{in}->{$inPos});
          }
          my $stat = Statistics::Descriptive::Full->new();
          $stat->add_data(@values);
          $Ratios{$type}->{$strand}->{$chr}->{$pos} = $stat->geometric_mean();
        }
        elsif( $mode eq "InLink"){       # ratio at position i is number of j's where there is a read-mate pair between i and j
          $Ratios{$type}->{$strand}->{$chr}->{$pos} = scalar(keys %{$data{$type}->{$strand}->{$chr}->{$pos}->{in}});
        }
        elsif( $mode eq "uniform"){
          # ratio at position i is coverages_i/(1+sum(abs((coverage_j-mean(coverage_j))/mean(coverage_j)))) for all j's where there is a read-mate pair between i and j
          # see http://stats.stackexchange.com/questions/25827/how-does-one-measure-the-non-uniformity-of-a-distribution
          my @values = ();
          foreach my $inPos (keys %{$data{$type}->{$strand}->{$chr}->{$pos}->{in}}){
            push @values, ( $data{$type}->{$strand}->{$chr}->{$pos}->{in}->{$inPos} );
          }
          my $mean = &arithmetic_mean(@values);
	  @values  = map {abs($_-$mean)/$mean} @values;
	  my $sum  = 0;
	  foreach my $value (@values){
	    $sum += $value;
	  }
          $Ratios{$type}->{$strand}->{$chr}->{$pos} = $data{$type}->{$strand}->{$chr}->{$pos}->{coverage}/(1+$sum);
        }
	elsif( $mode eq "coverage"){     # ratio at position i is coverage_i
          $Ratios{$type}->{$strand}->{$chr}->{$pos} = $data{$type}->{$strand}->{$chr}->{$pos}->{coverage}; 
        }
        else{
	  print STDERR "Error:\tno appropiate mode specified. Specify via -mode <InMax|GeomMean|InLink|uniform|coverage> a mode. If omitted <GeomMean> is used.\n";
	  exit;
        }
      }
    }
  }
}
print STDERR "\t\t..finished\n" if ($verbose);

# ranke control distribution and generate sorted array
@time=localtime();
$time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));
print STDERR "Log:\tStart sorting control @ $time_string\n" if ($verbose);
my @sortedControl = (0);
my %ratio_pvalue_map = ();
foreach my $type ($controlSample){
  foreach my $strand ( sort keys %{$Ratios{$type}} ){
    foreach my $chr ( sort keys %{$Ratios{$type}->{$strand}} ){
      foreach my $pos ( sort keys %{$Ratios{$type}->{$strand}->{$chr}} ){
        push @sortedControl, $Ratios{$type}->{$strand}->{$chr}->{$pos};
      }
    }
  }
}
@sortedControl = sort {$a <=> $b} @sortedControl;
print STDERR "\t\t..finished\n" if ($verbose);

# calculate empiric p-value on the fly and print
@time=localtime();
$time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));
print STDERR "Log:\tStart calculating p-values and printing results @ $time_string\n" if ($verbose);
print join("\t", qw/readType chrom chromStart chromEnd strand readCounts inSources Ratio pValue/)."\n";
foreach my $type (keys %Ratios){
  foreach my $strand ( sort keys %{$Ratios{$type}} ){
    foreach my $chr ( sort keys %{$Ratios{$type}->{$strand}} ){
      foreach my $pos ( sort keys %{$Ratios{$type}->{$strand}->{$chr}} ){
        my $pvalue     = ( $type eq $testSample )?(&incorpIntoSortedList_QD($Ratios{$type}->{$strand}->{$chr}->{$pos}, \@sortedControl)):(1);
        my $readCounts = $data{$type}->{$strand}->{$chr}->{$pos}->{coverage};
        my $sources    = scalar( keys %{$data{$type}->{$strand}->{$chr}->{$pos}->{in}} );
        my $ratio      = $Ratios{$type}->{$strand}->{$chr}->{$pos};
	
        print join("\t", $type, $chr, $pos-1, $pos, $strand, $readCounts, $sources, $ratio, $pvalue)."\n";
      }
    }
  }
}
print STDERR "\t\t..finished\n" if ($verbose);
@time=localtime();
$time_string = join(":", sprintf("%02d", $time[2]), sprintf("%02d", $time[1]), sprintf("%02d", $time[0]));
print STDERR "Log:\tAnalysis done @ $time_string\n" if ($verbose);

#################
## subroutines ##
#################

sub incorpIntoSortedList_QD{ # quick and dirty way to calculate p-value # use &incorpIntoSortedList for nice and slow one
  my $valueToSort = shift;
  my $sortedArray = shift;
  
  # this line makes it faster; sorting is not done for the 
  # smalles possible ratio but simply p-value =1 is reported
  if ( $valueToSort <= 100 ){
    if ( defined($ratio_pvalue_map{$valueToSort}) ){
      return $ratio_pvalue_map{$valueToSort};
    }
    else{
      my $lastIndex = List::MoreUtils::last_index {$_ < $valueToSort} @$sortedArray;
      $lastIndex = 0 if ($lastIndex == -1);
      my $pvalue = 1-($lastIndex/scalar(@$sortedArray));
      $ratio_pvalue_map{$valueToSort} = $pvalue;
      return($pvalue);
    }
  }
  else{
    my $lastIndex = List::MoreUtils::last_index {$_ < $valueToSort} @$sortedArray;
    $lastIndex = 0 if ($lastIndex == -1);
    return(1-($lastIndex/scalar(@$sortedArray)));
  }
}


sub incorpIntoSortedList{
  my $valueToSort = shift;
  my $sortedArray = shift;
  
  # this line makes it faster; sorting is not done for the 
  # smallest possible ratio but simply p-value =1 is reported
  if ( $valueToSort <= 1 ){
    return(1);
  }
  
  my $lastIndex = List::MoreUtils::last_index {$_ < $valueToSort} @$sortedArray;
  $lastIndex = 0 if ($lastIndex == -1);
  return(1-($lastIndex/scalar(@$sortedArray)));
}

sub cigarlength {
  my $cigar_string = shift;
  my $cigar_length = 0;
  
  while($cigar_string=~m/(\d+)[MDX=N]/g){
    $cigar_length+=$1;
  }
  
  if($cigar_length == 0) {
    print STDERR "Warning:\tfrom &cigarlength: CIGAR string <'$cigar_string'> seems corrupt;\n";
    return(0);
  }
  return($cigar_length);
}

sub arithmetic_mean {
  my @val = @_;
  
  my $sum = 0;
  my $n   = 0;
  
  if(scalar(@val)<1){
    return('NaN');
  }
  
  foreach my $val (@val){
    $sum += $val;
    $n++;
  }
  if ($n > 0 ){
    return($sum/$n);
  }
  else{
    return('NaN');
  }
}

sub log_mean {
  my @val = @_;
  
  my $sum = 0;
  my $n   = 0;
  
  if(scalar(@val)<1){
    return('NaN');
  }
  
  foreach my $val (@val){
    $sum += log($val);
    $n++;
  }
  if ($n > 0 ){
    return(exp($sum/$n));
  }
  else{
    return('NaN');
  }
}

##############
## man page ##
##############

=pod
    
=head1 NAME

-- EndEnrichmentInternalControl.pl -- 

=head1 SYNOPSIS

EndEnrichmentInternalControl.pl -bam I<STRING> [-mode I<STRING>] [-mr I<INT>] [-mf I<INT>] [-bg I<STRING>] [-v]  [--help|?] [--man] > OUTPUT.tsv

=head1 DESCRIPTION

Algorithm to detect transcription ends in paired-end term-seq data. Term-seq enriches cDNA associated with primary 3' ends of transcripts by ligating 3' specific adapter sequences prior to RNA fragmentation. After fragmentation true transcript 3' ends are expected to be represented by multiple independent cDNA fragments which end at the very same position (the true 3' end of the transcript) but starting at different 5' starts (which are randomly generated during RNA fragmentation).

To extract such genomic position from the data we make use that 3' ends and 5' ends of the same fragment is determined precisely in paired-end data. We are looking for positions where many reads end whose associated mates start at different positions. To this end we developed several measures quantifying the concentration of mate start positions onto a read end position. To evaluate enrichment and quantify an expectation value the distribution of concentration of read start positions onto mate start positions is used as the background. 

=head1 OPTIONS

=over 4

=item B<-bam> I<STRING>

Path to input bam file (must be paired end data). Obligatory.

=item B<-mode> I<InMax|GeomMean|InLink|uniform|coverage>

Mode describes the scoring scheme applied to capture the concentration of different mate start positions I<j> onto read start position I<i>. Default is I<GeomMean>.

=over 8

=item B<InMax> Ratio at position I<i> is I<coverage_i/max(coverages_j)> for all I<j>'s where there is a read-mate pair between I<i> and I<j>.

=item B<GeomMean> Ratio at position I<i> is I<geometric_mean(coverage_i/coverages_j)> for all I<j>'s where there is a read-mate pair between I<i> and I<j>.

=item B<InLink> Ratio at position I<i> is number of I<j>'s where there is a read-mate pair between I<i> and I<j>.

=item B<uniform> Ratio at position I<i> is I<coverages_i/(1+sum(abs((coverage_j-mean(coverage_j))/mean(coverage_j))))> for all I<j>'s where there is a read-mate pair between I<i> and I<j>.

=item B<coverage> Score at position I<i> is I<coverage_i>.

=back 

=item B<-mf> I<INT>

Only proper pairs with a insert length less then I<INT> are considered. Default is I<1000>.

=item B<mr> I<INT>

Only positions with at least I<INT> reads ending there are considered as true fragment termini. Default is I<1>.

=item B<-bg> I<STRING>

Path to file into which read end coverage of reads passing -mr and -mf filter are written (in bedgraph format). Default I<coverage.bg>. Existing files with this name will be overwritten.

=item B<-samtools> I<STRING>

Path to the samtools program. Default I<samtools> assuming it is available via the $PATH variable.

=item B<--verbose>

If set some progress reports are printed to STDERR during computation.

=item B<--man>

Print the long version of the man-page.

=item B<--[help|?]>

Print the short version of the man-page.

=back

=head1 OUTPUT FILE

Output is written on STDOUT in the form of a <TAB>-separated file with 9 columns, containing the following information:

=over 8

=item B<readType:>

I<mate> or I<read>. Typically I<read> is assumed to be the background distrubtion (no enrichment due to the library preparation protocol) and read is assumed to be the enriched signal. If this for some reason is flipped in your set-up, please consider to reset the variables $controlSample and $testSample directly in the code.

=item B<chrom:>

Chromosome name extracted from the bam file.

=item B<chromStart:>

Considered position (0-based).

=item B<chromEnd:>

Considered position (1-based).

=item B<strand:>

Strand of the considered position (rev: for - strand; fwd for + strand).

=item B<readCounts:>

Number of read or mate ends, respectively, at the considered positions.

=item B<inSources:>

Number of mate or read end positions associated with this rate or mate position considered.

=item B<Ratio:>

Score of the considered position (depending on the mode used).

=item B<pValue:>

Set to one for <mates>. Set for each score of <reads> to the percentile of this score in the distribution of <mate>-scores.

=back

=head1 VERSION

Version 0.9 beta

=head1 AUTHOR

Fabian Amman, fabian@tbi.univie.ac.at

=head1 LICENCE

GPL (>= 2)

=cut
